#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "usart.h"
#include "adc.h"
#include "lcd.h"
#include <math.h>
#define SOUND_FREQUENCY		50

char buf[50];
int pulsePin = 0;                 			    // Pulse Sensor connected to pin A0
volatile int rate[10];                    	// array to hold last ten IBI values
volatile unsigned long sampleCounter = 0;  	// used to determine pulse timing
volatile unsigned long lastBeatTime = 0;    // used to find IBI
volatile int P = 512;                      	// used to find peak in pulse wave, seeded
volatile int T = 512;                     	// used to find trough in pulse wave, seeded
volatile int thresh = 530;                	// used to find instant moment of heart beat, seeded
volatile int amp = 100;                   	// used to hold amplitude of pulse waveform, seeded
volatile bool firstBeat = true;     // used to seed rate array so we startup with reasonable BPM
volatile bool secondBeat = false;   // used to seed rate array so we startup with reasonable BPM
volatile int BPM;                   // int that holds raw Analog in 0. updated every 2mS
volatile int Signal;                // holds the incoming raw data
volatile int IBI = 600;             // holds the time interval between beats! Must be seeded!
volatile bool Pulse = false;     // "True" when User's live heartbeat is detected. "False" when not a "live beat"
volatile bool QS = false;        // becomes true when a beat is detected


void turn_on_speaker(int frequency) {
	int i;

	/* Alternate PD5 pin status every milisecond */
	for (i = 0; i < frequency; ++i) {
		PIND = (1 << PD5);

		_delay_ms(1);
	}
}

void turn_off_leds(void) {
  PORTB = 0x00;
}

/* Turn on the led according with BPM value */
void turn_on_led(void) {
	if(BPM >= 60 && BPM <= 100) {
      turn_off_leds();
      PORTB ^= (1 << PB2);  // turn on the green LED
  }
  else if( (BPM < 60 && BPM >= 50) || (BPM > 100 && BPM <= 120)){
      turn_off_leds();
      PORTB ^= (1 << PB1);  // turn on the yellow LED
  }
  else{
      turn_off_leds();
      PORTB ^= (1 << PB0);    // turn on the red LED
  }
}

/* Interruption of comparison with OCR1A */
ISR(TIMER1_COMPA_vect){                                                            
	cli();                      			       // disable interrupts while we do this
  Signal = ADC_get(pulsePin);         	   // read signal from the Pulse Sensor  
  sampleCounter += 2;    					         // keep track of the time in mS
  int N = sampleCounter - lastBeatTime;    // monitor the time since the last beat to avoid noise

  // find the peak and trough of the pulse wave
	if(Signal < thresh && N > (IBI / 5) * 3){   // avoid dichrotic noise by waiting 3/5 of last IBI
	    if (Signal < T){                        // T is the trough
	      T = Signal;                         	// keep track of lowest point in pulse wave
	    }
	  }

  	if(Signal > thresh && Signal > P){        // thresh condition helps avoid noise
    	P = Signal;                             // P is the peak
  	}                                        	// keep track of highest point in pulse wave

  // look for the heartbeats
  if (N > 250){                             // wait 250 ms to avoid high frequency noise
    if ( (Signal > thresh) && (Pulse == false) && (N > (IBI / 5) * 3) ){
    	Pulse = true;                         // set the Pulse flag when we think there is a pulse
	    //turn_off_leds();         			        
	    turn_on_speaker(SOUND_FREQUENCY);		  // turn on buzzer
	    IBI = sampleCounter - lastBeatTime;   // measure time between beats in mS
	    lastBeatTime = sampleCounter;         // keep track of time for next pulse
	    
	    if(secondBeat){                        // if this is the second beat
	        secondBeat = false;                // clear secondBeat flag
	        for(int i = 0; i <= 9; i++){       // seed the running total to get a realisitic BPM at startup
	          rate[i] = IBI;
	        }
	    }

	    if(firstBeat){                         // if it's the first time we found a beat
	        firstBeat = false;                 // clear firstBeat flag
	        secondBeat = true;                 // set the second beat flag
	        sei();                             // enable interrupts again
	        return;                            // IBI value is unreliable so discard it
	    }


      // collect the last 10 IBI values
   		int runningTotal = 0;                  // clear the runningTotal variable

  		for(int i = 0; i <= 8; i++){             // shift data in the rate array
          	rate[i] = rate[i + 1];             // and drop the oldest IBI value
          	runningTotal += rate[i];           // add up the 9 oldest IBI values
      }

  		rate[9] = IBI;                          // add the latest IBI to the rate array
      runningTotal += rate[9];                // add the latest IBI to runningTotal
      runningTotal /= 10;                     // average the last 10 IBI values
        	
      // BPM = how many beats can fit into a minute
      BPM = 60000/runningTotal;
      QS = true;                              // set Quantified Self flag
    }
  }

  if (Signal < thresh && Pulse == true){   	// when the values are going down, the beat is over
	  //turn_off_leds();                        // turn off leds
    Pulse = false;                         	// reset the Pulse flag so we can do it again
    amp = P - T;                           	// get amplitude of the pulse wave
    thresh = amp/2 + T;                    	// set thresh at 50% of the amplitude
    P = thresh;                           	// reset these for next time
    T = thresh;
  }

  if (N > 2500){                           // if 2.5 seconds go by without a beat
    thresh = 530;                          // set thresh default
    P = 512;                               // set P default
    T = 512;                               // set T default
    lastBeatTime = sampleCounter;          // bring the lastBeatTime up to date
    firstBeat = true;                      // set these to avoid noise
    secondBeat = false;                    // when we get the heartbeat back
  }

  sei();                                   // enable interrupts when youre done!                         
}

/* Init of timer 1 */
void timer1_init()
{
	// CTC mode with top at OCR1A = 124, 0x7C, value for 2ms, PS = 256
	TCCR1A = 0x00;
  TCCR1B = 0x0C; 
  OCR1A = 0x7C; 
  TIMSK1 = 0x02;  // enable interrupt for comparision with OCR1A

  sei(); // enable global interrupts
}

void IO_init(){
	/* Set pins PB0, PB1, PB2 at output*/
	DDRB |= (1 << PB0);  
	DDRB |= (1 << PB1);
	DDRB |= (1 << PB2);

	 /* Set pin PD5 at output for speaker*/
	DDRD |= (1 << PD5);

	 /* Turn off speaker*/
	PORTD &= ~(1 << PD5);

  turn_off_leds();
}

int main() 
{ 	
	//Initializations
	USART0_init();
	ADC_init();
	timer1_init();
  IO_init();
  LCD_init();

  // Backlight LCD
  DDRC |= (1 << PC2);
  PORTC |= (1 << PC2);

    while(1){

  		if (QS == true){     // A Heartbeat was found, BPM has been determined
  			turn_on_led();

        sprintf(buf, "BPM = %d\n", BPM); // send to USART
  	  	USART0_print(buf);

        LCD_clear_display(); // send to LCD
        sprintf(buf, "BPM = %d", BPM);
        LCD_printAt(0, buf); 

          	
          	
          	QS = false;    // reset the Quantified Self flag for next time
  		}	

  	  _delay_ms(20);      //  take a break
  }	 
	return 0;
}
